from cffi.api import FFI
import os
from cffi.recompiler import recompile, _modname_to_file, _unpatch_meths
from cffi import ffiplatform

def generate_c(ffi):
    module_name, source, source_extension, kwds = ffi._assigned_source
    recompile(ffi, module_name, source, call_c_compiler=False
              , source_extension=source_extension,
              compiler_verbose=True)
    tmpdir = '.'
    if not isinstance(module_name, str):
        module_name = module_name.encode('ascii')
    if ffi._windows_unicode:
        ffi._apply_windows_unicode(kwds)
    embedding = False
    c_file, parts = _modname_to_file(tmpdir, module_name,
                                     source_extension)
    ext_c_file = os.path.join(*parts)
    original_ending = '''#ifdef __GNUC__
#  pragma GCC visibility push(default)  /* for -fvisibility= */
#endif

#ifdef PYPY_VERSION
PyMODINIT_FUNC
_cffi_pypyinit_'''+ module_name +'''(const void *p[])
{
    p[0] = (const void *)0x2601;
    p[1] = &_cffi_type_context;
#if PY_MAJOR_VERSION >= 3
    return NULL;
#endif
}
#  ifdef _MSC_VER
     PyMODINIT_FUNC
#  if PY_MAJOR_VERSION >= 3
     PyInit_''' + module_name + '''(void) { return NULL; }
#  else
     init'''+ module_name + '''(void) { }
#  endif
#  endif
#elif PY_MAJOR_VERSION >= 3
PyMODINIT_FUNC
PyInit_'''+ module_name + '''(void)
{
  return _cffi_init("'''+ module_name +'''", 0x2601, &_cffi_type_context);
}
#else
PyMODINIT_FUNC
init'''+ module_name + '''(void)
{
  _cffi_init("'''+ module_name +'''", 0x2601, &_cffi_type_context);
}
#endif

#ifdef __GNUC__
#  pragma GCC visibility pop
#endif'''
    new_ending = '''#ifdef __GNUC__
#  pragma GCC visibility push(default)  /* for -fvisibility= */
#endif

static void sprintf_hex(const unsigned char *buffer, int len, const char *out){
        int i;
        char *temp=out;
        for(i = 1; i <= len; i++){
                sprintf(temp, "%02x",buffer[i-1]);
                temp += 2;
        }
}

static void sha128 (char * message, size_t message_len, char * digest_value_hex) {
    unsigned int digest_length = 0;
    unsigned char digest_value[21];
    if (sha1((unsigned char *)message, message_len, digest_value, &digest_length) == 1) {
        if (digest_length > 20) {
                PyErr_SetString(PyExc_RuntimeError,"Checksum failed5");
        }
        sprintf_hex(digest_value, digest_length, digest_value_hex);
    }
    else {
        PyErr_SetString(PyExc_RuntimeError,"Checksum failed4");
    }
}

typedef struct file_content {
    void * content;
    int length;
} file_content;

static file_content readfile(int filedescriptor) {
    #define SHARED_ATOMIC_FILE_BLOCK_SIZE 2097152
    char buffer [SHARED_ATOMIC_FILE_BLOCK_SIZE];
    #ifdef _MSC_VER
        char * content = (char *)malloc(SHARED_ATOMIC_FILE_BLOCK_SIZE);
    #else
        char * content = (char *)valloc(SHARED_ATOMIC_FILE_BLOCK_SIZE);
    #endif
    int i=1;
    int length=0;
    int read_num_bytes;
    while((read_num_bytes = read(filedescriptor, buffer, SHARED_ATOMIC_FILE_BLOCK_SIZE))) {
        memcpy(content + (i-1) * SHARED_ATOMIC_FILE_BLOCK_SIZE, buffer, read_num_bytes);
        length += read_num_bytes;
        if (read_num_bytes < SHARED_ATOMIC_FILE_BLOCK_SIZE) {
            break;
        }
        else {
            content = realloc(content, SHARED_ATOMIC_FILE_BLOCK_SIZE * (++i));
        }
    };

    struct file_content result={content, length};
    return  result;
}

static _Bool checksum() {
    char digest_value_hex [41];
    sha128("===randomstring===", strlen("===randomstring==="), digest_value_hex);
    if (strcmp(digest_value_hex, "===randomstringsha1===") != 0) {
        fprintf(stderr, "Checksum failed1!\\n");
        return false;
    };

    //Py_Initialize();
    PyObject *pName,*pModule, *pFunc, *result;
    pName = PyUnicode_DecodeFSDefault("shared_atomic.atomic_activation");
    pModule = PyImport_Import(pName);
    char * file_path=PyModule_GetFilename(pModule);

    #ifdef _MSC_VER
    int file_path_length = MultiByteToWideChar(CP_UTF8,MB_PRECOMPOSED,file_path,-1,NULL,0);
    LPWSTR file_path_wchar = (LPWSTR) malloc(file_path_length * sizeof(WCHAR));
    file_path_length = MultiByteToWideChar(CP_UTF8,MB_PRECOMPOSED,file_path,-1,file_path_wchar,file_path_length);

    if ( file_path_length <= 0) {
        fprintf(stderr, "Checksum failed4!\\n");
        return false;
    }

    HANDLE file_handle = CreateFileW(
			file_path_wchar,
			GENERIC_READ,
			FILE_SHARE_READ,
			NULL,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
			NULL);

	DWORD filesize = GetFileSize(file_handle,NULL);
	
	if ( filesize == INVALID_FILE_SIZE) {
	    fprintf(stderr, "Checksum failed7!\\n");
        return false;
	}

	HANDLE hMapFileVTX = CreateFileMappingW(file_handle,NULL, PAGE_READONLY, 0, 0, NULL);
	
	if ( hMapFileVTX == NULL) {
	    fprintf(stderr, "Checksum failed8!\\n");
        return false;
	}
	
	LPVOID lpBase = MapViewOfFile(hMapFileVTX, FILE_MAP_READ, 0, 0, 0);
	
	if ( lpBase == NULL) {
	    fprintf(stderr, "Checksum failed9!\\n");
        return false;
	}

    free(file_path_wchar);
    struct file_content file = {lpBase, filesize};

    #else
    int fd;
    if( (fd = open(file_path, O_RDONLY)) == -1 ) {
        fprintf(stderr, "Checksum failed2!\\n");
        return false;
    }


    struct file_content file = readfile(fd);
    #endif

    sha128(file.content, file.length, digest_value_hex);

    #ifdef _MSC_VER
    if ( UnmapViewOfFile(lpBase) == 0) {
        fprintf(stderr, "Checksum failed6!\\n");
        return false;
    }
    if ( CloseHandle(file_handle) == 0) {
        fprintf(stderr, "Checksum failed5!\\n");
        return false;
    }
    #else
    free(file.content);
    close(fd);
    #endif

    if (strcmp(digest_value_hex,"===activationlibsha1===") != 0) {
        fprintf(stderr, "Checksum failed3!\\n");
        return false;
    };

    pFunc= PyObject_GetAttrString(pModule, "verify_activation");
    result= PyObject_CallObject(pFunc, NULL);
    if (result == NULL) {
        return false;
    }
    else if (PyObject_IsTrue(result)) {
        return true;
    }
    else {
        return false;
    };

    //Py_Finalize();

}

#ifdef PYPY_VERSION
PyMODINIT_FUNC
_cffi_pypyinit_'''+ module_name +'''(const void *p[])
{
    if (checksum() == false)
        PyErr_SetString(PyExc_RuntimeError,"verify_activation failed");
    else {
        p[0] = (const void *)0x2601;
        p[1] = &_cffi_type_context;
    #if PY_MAJOR_VERSION >= 3
        return NULL;
    #endif
    }
}
    #ifdef _MSC_VER
         PyMODINIT_FUNC
        #if PY_MAJOR_VERSION >= 3
             PyInit_'''+ module_name +'''(void) { 
                if (checksum() == false) 
                    PyErr_SetString(PyExc_RuntimeError,"verify_activation failed");
                else
                    return NULL; 

             }
        #else
             init'''+ module_name +'''(void) { 
                if (! checksum()) 
                    PyErr_SetString(PyExc_RuntimeError,"verify_activation failed");
             }
        #endif
    #endif
#elif PY_MAJOR_VERSION >= 3
    PyMODINIT_FUNC
    PyInit_'''+ module_name +'''(void)
    {
      if ( checksum() == false) {
            PyErr_SetString(PyExc_RuntimeError,"verify_activation failed");
      }
      else {
            return _cffi_init("'''+ module_name +'''", 0x2601, &_cffi_type_context);
      }

    }
#else
    PyMODINIT_FUNC
    init'''+ module_name +'''(void)
    {
        if (checksum() == false) 
            PyErr_SetString(PyExc_RuntimeError,"verify_activation failed");
        else
            _cffi_init("'''+ module_name +'''", 0x2601, &_cffi_type_context);
    }
#endif

#ifdef __GNUC__
#  pragma GCC visibility pop
#endif'''

    cffi_dependencies = 'PyImport_ImportModule("_cffi_backend")'
    cffi_shared_atomic_dependencies = 'PyImport_ImportModule("shared_atomic._cffi_backend")'
    with open(ext_c_file, 'rt') as f:
        ext_c_file_content = f.read()
    with open(ext_c_file, 'wt') as f:
        #print("original_ending: " + original_ending)
        #print("new_ending: "+new_ending)
        f.write(ext_c_file_content.replace(original_ending, new_ending)
                .replace(cffi_dependencies,cffi_shared_atomic_dependencies))
    return ext_c_file

def compile(ffi):
    module_name, source, source_extension, kwds = ffi._assigned_source
    tmpdir = '.'
    if not isinstance(module_name, str):
        module_name = module_name.encode('ascii')
    if ffi._windows_unicode:
        ffi._apply_windows_unicode(kwds)
    c_file, parts = _modname_to_file(tmpdir, module_name,
                                     source_extension)
    ext_c_file = os.path.join(*parts)
    ext = ffiplatform.get_extension(ext_c_file, module_name, **kwds)
    patchlist = []
    cwd = os.getcwd()
    try:
        if tmpdir == '.':
            msg = 'the current directory is'
        else:
            msg = 'setting the current directory to'
        print('%s %r' % (msg, os.path.abspath(tmpdir)))
        os.chdir(tmpdir)
        outputfilename = ffiplatform.compile('.', ext,
                                             True, False)
    finally:
        os.chdir(cwd)
        _unpatch_meths(patchlist)
    return outputfilename
