'''
Created on 28 Mar 2022

@author: philren
'''

import sys
import subprocess
import re
from pathlib import Path
import cffi
import os
from cffi.recompiler import recompile, _modname_to_file, _unpatch_meths
import shared_atomic.atomic_cffi_compile_steps as steps
import distutils

if sys.platform not in ('linux', 'darwin','win32'):
    raise NotImplementedError("Unsupported platform!")

ffi = cffi.FFI()

hfilepath = Path.joinpath(Path(__file__).parent, 'atomic_decryption.h')

if sys.platform in ('linux', 'darwin'):
    header = '''int flock(int fd, int operation);
void *malloc(size_t size);
void free(void *ptr);
void *realloc(void *ptr, size_t size);
'''
else:
    header = '''
    void * CreateFileAtomic(wchar_t * lpFileName,
				unsigned long dwDesiredAccess,
				unsigned long dwShareMode,
				unsigned long dwCreationDisposition,
				unsigned long dwFlagsAndAttributes);

	int ReadFileAtomic(void *       hFile,
		void *       lpBuffer,
		unsigned long        nNumberOfBytesToRead,
		unsigned long *      lpNumberOfBytesRead);

    int WriteFileAtomic(void *       hFile,
		void *      lpBuffer,
		unsigned long        nNumberOfBytesToWrite,
		unsigned long *     lpNumberOfBytesWritten);

    int CloseHandleAtomic(void * hObject);
    int GetHardwareInfo(unsigned char * nonce, unsigned char * result_buffer, int * result_size);
    int GetSerialNumber(unsigned char * nonce, unsigned char * result_buffer, int * result_size);
    int TruncateFileAtomic(void * hFile);
'''

with hfilepath.open() as hfile:
        header += ''.join(hfile.readlines())

ffi.cdef(header)

if sys.platform in('darwin','linux'):
    result = subprocess.run('gcc -x c -v -E /dev/null'.split(),
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)

    result = bytes.decode(result.stderr).split("\n")

    start = result.index('#include <...> search starts here:')
    end = result.index('End of search list.')

    result_include = ["-I" + re.sub(r" *\(framework directory\)$", "", i).strip() for i in result[start + 1:end]] \
                     + ["-I" + str(Path.joinpath(Path(__file__).parent, "include"))]
    result_link = result_include + ["-L" + str(Path.joinpath(Path(__file__).parent, "lib").absolute())]
else:
    result_include = ["-I.\include"]
    result_link = result_include + ["/LIBPATH:" + str(Path(__file__).parent.absolute())]


if sys.platform in ('linux', 'darwin'):
    ffi.set_source('atomic_decryption_', """
        #include <stddef.h>
        #include <sys/types.h>
        #include <sys/mman.h>
        #include <time.h>
        #include <stdbool.h>
        #include <libgen.h>
        #include <sys/file.h>
        """ + header,
                   #sources=[str(cfilepath)],
                   extra_compile_args=result_include,
                   libraries=['decryption','decryption11'],
                   extra_link_args=result_link
                   )
else:
    cfilepath = Path.joinpath(Path(__file__).parent, 'atomic_decryption_win64.c')
    ffi.set_source('atomic_decryption_', '''
        #include <windows.h>
        #include <BaseTsd.h>
        #include <stdbool.h>
        #include <stdio.h>
        #include <string.h>
        #include <stringapiset.h>
        #include <fileapi.h>
        #include <memoryapi.h>
        #include <handleapi.h>
        ''' + header,
                   sources=[str(cfilepath)],
                   extra_compile_args=result_include,
                   libraries=['libdecryption','libdecryption11'],
                   extra_link_args=result_link
                   )


def generate_c():
    module_name, source, source_extension, kwds = ffi._assigned_source
    recompile(ffi, module_name, source, call_c_compiler=False
              , source_extension=source_extension,
              compiler_verbose=True)
    tmpdir = '.'
    if not isinstance(module_name, str):
        module_name = module_name.encode('ascii')
    if ffi._windows_unicode:
        ffi._apply_windows_unicode(kwds)
    embedding = False
    c_file, parts = _modname_to_file(tmpdir, module_name,
                                     source_extension)
    ext_c_file = os.path.join(*parts)
    original_ending = '''#ifdef __GNUC__
#  pragma GCC visibility push(default)  /* for -fvisibility= */
#endif

#ifdef PYPY_VERSION
PyMODINIT_FUNC
_cffi_pypyinit_''' + module_name + '''(const void *p[])
{
    p[0] = (const void *)0x2601;
    p[1] = &_cffi_type_context;
#if PY_MAJOR_VERSION >= 3
    return NULL;
#endif
}
#  ifdef _MSC_VER
     PyMODINIT_FUNC
#  if PY_MAJOR_VERSION >= 3
     PyInit_''' + module_name + '''(void) { return NULL; }
#  else
     init''' + module_name + '''(void) { }
#  endif
#  endif
#elif PY_MAJOR_VERSION >= 3
PyMODINIT_FUNC
PyInit_''' + module_name + '''(void)
{
  return _cffi_init("''' + module_name + '''", 0x2601, &_cffi_type_context);
}
#else
PyMODINIT_FUNC
init''' + module_name + '''(void)
{
  _cffi_init("''' + module_name + '''", 0x2601, &_cffi_type_context);
}
#endif

#ifdef __GNUC__
#  pragma GCC visibility pop
#endif'''
    new_ending = '''#ifdef __GNUC__
#  pragma GCC visibility push(default)  /* for -fvisibility= */
#endif

static void sprintf_hex(const unsigned char *buffer, int len, const char *out){
        int i;
        char *temp=out;
        for(i = 1; i <= len; i++){
                sprintf(temp, "%02x",buffer[i-1]);
                temp += 2;
        }
}

static void sha128 (char * message, size_t message_len, char * digest_value_hex) {
    unsigned int digest_length = 0;
    unsigned char digest_value[21];
    if (sha1((unsigned char *)message, message_len, digest_value, &digest_length) == 1) {
        if (digest_length > 20) {
                PyErr_SetString(PyExc_RuntimeError,"Checksum failed5");
        }
        sprintf_hex(digest_value, digest_length, digest_value_hex);
    }
    else {
        PyErr_SetString(PyExc_RuntimeError,"Checksum failed4");
    }
}

typedef struct file_content {
    void * content;
    int length;
} file_content;

static file_content readfile(int filedescriptor) {
    #define SHARED_ATOMIC_FILE_BLOCK_SIZE 2097152
    char buffer [SHARED_ATOMIC_FILE_BLOCK_SIZE];
    #ifdef _MSC_VER
        char * content = (char *)malloc(SHARED_ATOMIC_FILE_BLOCK_SIZE);
    #else
        char * content = (char *)valloc(SHARED_ATOMIC_FILE_BLOCK_SIZE);
    #endif
    int i=1;
    int length=0;
    int read_num_bytes;
    while((read_num_bytes = read(filedescriptor, buffer, SHARED_ATOMIC_FILE_BLOCK_SIZE))) {
        memcpy(content + (i-1) * SHARED_ATOMIC_FILE_BLOCK_SIZE, buffer, read_num_bytes);
        length += read_num_bytes;
        if (read_num_bytes < SHARED_ATOMIC_FILE_BLOCK_SIZE) {
            break;
        }
        else {
            content = realloc(content, SHARED_ATOMIC_FILE_BLOCK_SIZE * (++i));
        }
    };

    struct file_content result={content, length};
    return  result;
}

static void sha_file(char * file_path, char * digest_value_hex) {
    #ifdef _MSC_VER
    int file_path_length = MultiByteToWideChar(CP_UTF8,MB_PRECOMPOSED,file_path,-1,NULL,0);
    LPWSTR file_path_wchar = (LPWSTR) malloc(file_path_length * sizeof(WCHAR));
    file_path_length = MultiByteToWideChar(CP_UTF8,MB_PRECOMPOSED,file_path,-1,file_path_wchar,file_path_length);

    if ( file_path_length <= 0) {
        fprintf(stderr, "Checksum failed4!\\n");
        return;
    }

    HANDLE file_handle = CreateFileW(
			file_path_wchar,
			GENERIC_READ,
			FILE_SHARE_READ,
			NULL,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL,
			NULL);

	DWORD filesize = GetFileSize(file_handle,NULL);

	if ( filesize == INVALID_FILE_SIZE) {
	    fprintf(stderr, "Checksum failed7!\\n");
        return;
	}

	HANDLE hMapFileVTX = CreateFileMappingW(file_handle,NULL, PAGE_READONLY, 0, 0, NULL);

	if ( hMapFileVTX == NULL) {
	    fprintf(stderr, "Checksum failed8!\\n");
        return;
	}

	LPVOID lpBase = MapViewOfFile(hMapFileVTX, FILE_MAP_READ, 0, 0, 0);

	if ( lpBase == NULL) {
	    fprintf(stderr, "Checksum failed9!\\n");
        return;
	}

    free(file_path_wchar);
    struct file_content file = {lpBase, filesize};

    #else
    int fd;
    if( (fd = open(file_path, O_RDONLY)) == -1 ) {
        fprintf(stderr, "Checksum failed2!\\n");
        return;
    }


    struct file_content file = readfile(fd);
    #endif

    sha128(file.content, file.length, digest_value_hex);

    #ifdef _MSC_VER
    if ( UnmapViewOfFile(lpBase) == 0) {
        fprintf(stderr, "Checksum failed6!\\n");
        return;
    }
    if ( CloseHandle(file_handle) == 0) {
        fprintf(stderr, "Checksum failed5!\\n");
        return;
    }
    #else
    free(file.content);
    close(fd);
    #endif
}
static _Bool checksum() {
    char digest_value_hex [41];

    sha128("===randomstring===", strlen("===randomstring==="), digest_value_hex);
    if (strcmp(digest_value_hex, "===randomstringsha1===") != 0) {
        fprintf(stderr, "Checksum failed1!\\n");
        return false;
    };

    //Py_Initialize();
    PyObject *pName,*pModule;
    pName = PyUnicode_DecodeFSDefault("shared_atomic._cffi_backend");
    pModule = PyImport_Import(pName);
    char * file_path=PyModule_GetFilename(pModule);

    sha_file(file_path, digest_value_hex);
    if (strcmp(digest_value_hex,"===cffilibsha1===") != 0) {
        fprintf(stderr, "Checksum failed3!\\n");
        return false;
    };
    
    #ifndef _MSC_VER
        char * libffi_path = dirname(file_path);
        #ifdef __APPLE__
            strcat(libffi_path, "/lib/libffi.dylib");
        #endif
        #ifdef __linux__
            strcat(libffi_path, "/lib/libffi.so");
        #endif
        sha_file(libffi_path, digest_value_hex);

        if (strcmp(digest_value_hex,"===libffisha1===") != 0) {
            fprintf(stderr, "Checksum failed10!\\n");
            return false;
        };
    #endif
    
    return true;

}

#ifdef PYPY_VERSION
PyMODINIT_FUNC
_cffi_pypyinit_''' + module_name + '''(const void *p[])
{
    if (checksum() == false)
        PyErr_SetString(PyExc_RuntimeError,"Checksum failed");
    else {
        p[0] = (const void *)0x2601;
        p[1] = &_cffi_type_context;
    #if PY_MAJOR_VERSION >= 3
        return NULL;
    #endif
    }
}
    #ifdef _MSC_VER
         PyMODINIT_FUNC
        #if PY_MAJOR_VERSION >= 3
             PyInit_''' + module_name + '''(void) { 
                if (checksum() == false) 
                    PyErr_SetString(PyExc_RuntimeError,"Checksum failed");
                else
                    return NULL; 

             }
        #else
             init''' + module_name + '''(void) { 
                if (! checksum()) 
                    PyErr_SetString(PyExc_RuntimeError,"Checksum failed");
             }
        #endif
    #endif
#elif PY_MAJOR_VERSION >= 3
    PyMODINIT_FUNC
    PyInit_''' + module_name + '''(void)
    {
      if ( checksum() == false) {
            PyErr_SetString(PyExc_RuntimeError,"Checksum failed");
      }
      else {
            return _cffi_init("''' + module_name + '''", 0x2601, &_cffi_type_context);
      }

    }
#else
    PyMODINIT_FUNC
    init''' + module_name + '''(void)
    {
        if (checksum() == false) 
            PyErr_SetString(PyExc_RuntimeError,"Checksum failed");
        else
            _cffi_init("''' + module_name + '''", 0x2601, &_cffi_type_context);
    }
#endif

#ifdef __GNUC__
#  pragma GCC visibility pop
#endif'''

    cffi_dependencies = 'PyImport_ImportModule("_cffi_backend")'
    cffi_shared_atomic_dependencies = 'PyImport_ImportModule("shared_atomic._cffi_backend")'
    with open(ext_c_file, 'rt') as f:
        ext_c_file_content = f.read()
    with open(ext_c_file, 'wt') as f:
        # print("original_ending: " + original_ending)
        # print("new_ending: "+new_ending)
        f.write(ext_c_file_content.replace(original_ending, new_ending)
                .replace(cffi_dependencies, cffi_shared_atomic_dependencies))
    return ext_c_file

def compile():
    return steps.compile(ffi)

def main():
    #ffi.compile(verbose=True)

    ext_c_file = generate_c()
    outputfilename = compile()
    return outputfilename

if __name__ == "__main__":
    main()
