import os

from shared_atomic.atomic_int import *

import sys
from threading import Thread
import multiprocessing
from multiprocessing import Process
if sys.platform != 'win32':
    Process = multiprocessing.get_context("fork").Process



def setup_function():
    """
    pre function for pytest
    :return: None
    """
    # if sys.platform in ('darwin','linux'):
    #     dlltype = ctypes.CDLL
    #     os.chdir('/Users/philren/.local/share/virtualenvs/spark-examples--HrH57AW/lib/python3.6/site-packages')
    #     filename = 'shared_atomic.cpython-36m-darwin.so'
    # elif sys.platform == "win32":
    #     dlltype = ctypes.windll
    # else:
    #     return
    #atomic = ctypes.LibraryLoader(dlltype).LoadLibrary(filename)

def teardown_function():
    pass

def test_init():
    a = atomic_int(2**63-1)
    assert a.get() == 2**63-1
    a = atomic_int(-2**63)
    assert a.get() == -2**63

def test_int():
    if sys.platform in ("linux","darwin"):
        a = atomic_int(2**63-1)
        assert a.get() == 2**63-1
        a.set(-2**63)
        assert a.get() == -2**63
        b = atomic_int(2**63-2)
        result = a.int_compare_and_set(b, 1)
        assert result == False
        assert b.get() == a.get()
        result = a.int_compare_and_set(b, 1)

        assert result == True
        assert a.get() == 1

        b = atomic_int(2**63-2)
        a.int_store(b)

        assert a.value == b.value

        result = a.int_get_and_set(2**63-3)
        assert result == 2**63-2
        assert a.value == 2**63-3

        c = atomic_int(2**63-4)
        b = atomic_int(2**63-5)
        a.int_shift(b, c)
        assert a.value == 2**63-5
        assert c.value == 2**63-3

        assert a.int_add_and_fetch(1) == 2**63-4
        assert a.int_sub_and_fetch(1) == 2**63-5
        assert a.int_and_and_fetch(1111) == 2**63-5 & 1111
        assert a.int_or_and_fetch(1111) == (2**63-5 & 1111) | 1111
        assert a.int_xor_and_fetch(1111) == ((2**63-5 & 1111) | 1111) ^ 1111
        assert a.int_nand_and_fetch(1111) == ~((((2**63-5 & 1111) | 1111) ^ 1111) & 1111)
        a.set(2**63-4)
        assert a.int_fetch_and_add(1) == 2**63-4
        assert a.int_fetch_and_sub(1) == 2**63-3
        assert a.int_fetch_and_and(1111) == 2**63-4
        assert a.int_fetch_and_or(1111) == 2**63-4 & 1111
        assert a.int_fetch_and_xor(1111) == (2**63-4 & 1111) | 1111
        assert a.int_fetch_and_nand(1111) == ((2**63-4 & 1111) | 1111) ^ 1111
        assert a.value == ~((((2**63-4 & 1111) | 1111) ^ 1111) & 1111)

        a.value = 2**63-5

        assert int_add_and_fetch(a.reference, 1) == 2**63-4
        assert int_sub_and_fetch(a.reference, 1) == 2**63-5
        assert int_and_and_fetch(a.reference,1111) == 2**63-5 & 1111
        assert int_or_and_fetch(a.reference, 1111) == (2**63-5 & 1111) | 1111
        assert int_xor_and_fetch(a.reference, 1111) == ((2**63-5 & 1111) | 1111) ^ 1111
        assert int_nand_and_fetch(a.reference, 1111) == ~((((2**63-5 & 1111) | 1111) ^ 1111) & 1111)
        a.set(2**63-4)
        assert int_fetch_and_add(a.reference,1) == 2**63-4
        assert int_fetch_and_sub(a.reference,1) == 2**63-3
        assert int_fetch_and_and(a.reference,1111) == 2**63-4
        assert int_fetch_and_or(a.reference,1111) == 2**63-4 & 1111
        assert int_fetch_and_xor(a.reference,1111) == (2**63-4 & 1111) | 1111
        assert int_fetch_and_nand(a.reference,1111) == ((2**63-4 & 1111) | 1111) ^ 1111
        assert a.value == ~((((2**63-4 & 1111) | 1111) ^ 1111) & 1111)

        a = atomic_int(1024)
        result = a.int_compare_and_set_value(1026, 1025)
        assert result == 1024
        result = a.int_compare_and_set_value(1026, 1024)
        assert result == 1024
        assert a.value == 1026

    else:
        a = atomic_int(2 ** 63 - 1)
        assert a.get() == 2 ** 63 - 1
        a.set(-2 ** 63)
        assert a.get() == -2 ** 63
        result = a.int_get_and_set(2 ** 63 - 1)
        assert result == -2 ** 63
        assert a.get() == 2 ** 63 - 1
        result = a.int_compare_and_set_value(2 ** 63 - 3, 2 ** 63 - 2)
        assert result == 2 ** 63 - 1
        result = a.int_compare_and_set_value(2 ** 63 - 3, 2 ** 63 - 1)
        assert result == 2 ** 63 - 1
        assert a.get() == 2 ** 63 - 3
        a = atomic_int(-2 ** 62)
        result = a.int_add_and_fetch(2 * (2 ** 62) - 1)
        assert result == 2 ** 62 - 1
        assert a.get() == 2 ** 62 - 1
        result = a.int_sub_and_fetch(2 * (2 ** 62) - 1)
        assert result == - 2 ** 62
        assert a.get() == - 2 ** 62
        result = a.int_fetch_and_add(2 * (2 ** 62) - 1)
        assert result == - 2 ** 62
        assert a.get() == 2 ** 62 - 1
        result = a.int_fetch_and_sub(2 * (2 ** 62) - 1)
        assert result == 2 ** 62 - 1
        assert a.get() == - 2 ** 62
        a.set(- 2 ** 61)
        result = a.int_fetch_and_and(2 ** 62)
        assert result == - 2 ** 61
        assert a.get() == (- 2 ** 61) & (2 ** 62)
        result = a.int_fetch_and_or(2 ** 62-1)
        assert result == 4611686018427387904 # (- 2 ** 63) & (2 ** 63)
        assert a.get() == 4611686018427387904 | (2 ** 62-1)
        result = a.int_fetch_and_xor(2 ** 62)
        assert result == 4611686018427387904 | (2 ** 62-1)
        assert a.get() == ((4611686018427387904 | (2 ** 62-1))) ^ (2 ** 62)

        a = atomic_int(((4611686018427387904 | (2 ** 62-1))) ^ (2 ** 62))
        assert a.int_bittest_and_set(1) == False
        assert '0'*(64-len(bin(a.value)[2:]))+bin(a.value)[2:] \
         == '0111111111111111111111111111111111111111111111111111111111111111'
        assert a.int_bittest_and_reset(1) == True
        assert '0'*(64-len(bin(a.value)[2:]))+bin(a.value)[2:] \
         == '0011111111111111111111111111111111111111111111111111111111111111'

def thread_run(i):
    int_add_and_fetch(i.reference, 100)

def test_thread_atomic():
    """
    test single process multiple threads
    :return: None
    """
    a = atomic_int(0)

    threadlist = []

    for i in range(10000):
        threadlist.append(Thread(target=thread_run, args=(a,)))

    for i in range(10000):
        threadlist[i].start()

    for i in range(10000):
        threadlist[i].join()

    assert a.value == 100 * 10000


if sys.platform != "win32":

    from shared_atomic.atomic_object import int_add_and_fetch
    def process_run(a):
        def subthread_run(a):
            int_add_and_fetch(a.reference, 100)

        threadlist = []
        for t in range(5000):
            threadlist.append(Thread(target=subthread_run, args=(a,)))

        for t in range(5000):
            threadlist[t].start()

        for t in range(5000):
            threadlist[t].join()

    def test_process_atomic():
        """
        test multiple processes
        :return: None
        """
        a = atomic_int(0, mode= 'm', windows_unix_compatibility=True)

        processlist = []
        for i in range(2):
            processlist.append(Process(target=process_run, args=(a,)))

        for i in range(2):
            processlist[i].start()

        for i in range(2):
            processlist[i].join()

        assert a.get() == 100*10000
else:
    def process_run_compatibility(a):
        def subthread_run(a):
            a.int_add_and_fetch(100)
        threadlist = []
        for t in range(50000):
            threadlist.append(Thread(target=subthread_run, args=(a,)))
        for t in range(50000):
            threadlist[t].start()
        for t in range(50000):
            threadlist[t].join()

    def test_process_atomic_compatibility():
        """
        test multiple processes
        :return: None
        """
        a = atomic_int(0, mode='m',windows_unix_compatibility=True)
        processlist = []
        for i in range(10):
            processlist.append(Process(target=process_run_compatibility, args=(a,)))
        for i in range(10):
            processlist[i].start()
        for i in range(10):
            processlist[i].join()
        assert a.get() == 100*10*50000

    def process_run(a):
        def subthread_run(reference):
            int_add_and_fetch(reference,100)

        reference = get_reference(a)

        threadlist = []
        for t in range(50000):
            threadlist.append(Thread(target=subthread_run, args=(reference.reference,)))

        for t in range(50000):
            threadlist[t].start()

        for t in range(50000):
            threadlist[t].join()

        release_reference(reference)



    def test_process_atomic_incompatible():
        """
        test multiple processes
        :return: None
        """
        a = atomic_int(0, mode='m')
        processlist = []

        for i in range(10):
            processlist.append(Process(target=process_run, args=(a,)))

        for i in range(10):
            processlist[i].start()

        for i in range(10):
            processlist[i].join()

        assert int_get_and_set(a.reference, 0) == 10*50000*100
