import traceback

from shared_atomic.atomic_string import atomic_string
from shared_atomic.atomic_int import atomic_int
import ctypes
import sys
from threading import Thread
import multiprocessing
from multiprocessing import Process
if sys.platform != 'win32':
    Process = multiprocessing.get_context("fork").Process


inlist = (
         'a',
         'é',
         '重1',
         '重重',
)


exlist = ('b',
         'è',
         '启2',
         '轻轻',
          )


def signed2unsigned(input, i):
    if input < 0:
        return int.to_bytes(input + 2**((i+1)*8),length=i+1, byteorder='big').lstrip(b'\0')
    return int.to_bytes(input,length=2**i, byteorder='big').lstrip(b'\0')


def setup_function():
    """
    pre function for pytest
    :return: None
    """

    # if sys.platform in ('darwin','linux'):
    #     dlltype = ctypes.CDLL
    #     os.chdir('/Users/philren/.local/share/virtualenvs/spark-examples--HrH57AW/lib/python3.6/site-packages')
    #     filename = 'shared_atomic.cpython-36m-darwin.so'
    # elif sys.platform == "win32":
    #     dlltype = ctypes.windll
    # else:
    #     return
    #atomic = ctypes.LibraryLoader(dlltype).LoadLibrary(filename)

def teardown_function():
    pass

def test_init():
    a = atomic_string('ab1234567', length=7)
    a.get_string()
    a = atomic_string('ab')
    assert a.value == 'ab'
    a = atomic_string('ab', length=7, paddingdirection='l', paddingstr='012')
    assert a.value == '01201ab'
    a = atomic_string('ab', length=7, paddingdirection='r', paddingstr='012')
    assert a.value == 'ab01201'
    a = atomic_string('ab', length=7)
    assert a.value == 'ab     '
    a = atomic_string('ab1234567', length=7, trimming_direction='l')
    assert a.value == '1234567'
    a = atomic_string('ab1234567', length=7)
    assert a.value == 'ab12345'


def test_resize():
    a = atomic_string('ab1234567', length=7)
    a.resize(6)
    assert a.value == 'ab1234'
    a.resize(5, trimming_direction='l')
    assert a.value == 'b1234'
    a.resize(7, paddingstr='a', paddingdirection='l')
    assert a.value == 'aab1234'
    a.resize(6, trimming_direction='l')
    assert a.value == 'ab1234'
    a.resize(7, paddingstr='a', paddingdirection='r')
    assert a.value == 'ab1234a'

def test_string():
    a = atomic_string('ab', length=7, paddingdirection='r', paddingstr='012', mode='m')
    assert a.get_string() == 'ab01201'
    a.set_string('abc')

    aa = atomic_string('aa', length=7, paddingdirection='r', paddingstr='012', mode='m')
    bb = atomic_string('bb', length=7, paddingdirection='r', paddingstr='012', mode='m')

    if sys.platform != 'win32':
        result = aa.string_compare_and_set(bb, 'ab')

        assert result == False
        assert bb.get_string() == aa.get_string()

        result = aa.string_compare_and_set(bb, 'ab')

        assert result == True
        assert aa.get_string() == 'ab'


def test_value_string():
    """
    test single process single thread
    :return: None
    """
    i = 0
    result = None
    try:
        for i in range(4):
            print("i=",f'{i}')
            if sys.platform in ('darwin','linux'):
                if i != 3:
                    a = atomic_string('a'*(2**i))
                else:
                    a = atomic_string('a'*7)

                result = []
                b = atomic_string(inlist[i])
                a.string_store(b)
                assert a.get_string() == inlist[i]
                result.append(a.string_get_and_set(exlist[i]))
                assert result[-1] == inlist[i]
                assert a.get_string() == exlist[i]
                c = []
                c.append(atomic_string(exlist[i]))
                result.append(a.string_compare_and_set(c[-1], inlist[i]))
                assert result[-1]
                assert a.get_string() == inlist[i]
                c.append(atomic_string(inlist[i]))
                c.append(atomic_string(exlist[i]))
                result.append(a.string_shift(c[-1], c[-2]))
                assert c[-1].get_string() == a.get_string()
                assert c[-2].get_string() == inlist[i]

                value = a.get_string()
                a.reencode('utf-16-le')
                assert a.value == value

                value = a.get_string()
                a.change_mode('m')
                assert value == a.get_string()
                assert a.mode == 'm'
                a.change_mode('s')
                assert value == a.get_string()
                assert a.mode == 's'
            else:
                if i != 3:
                    a = atomic_string('a' * (2 ** i))
                else:
                    a = atomic_string('a' * 7)

                result = []
                a = atomic_string(inlist[i])
                assert a.get_string() == inlist[i]
                result.append(a.string_get_and_set(exlist[i]))
                assert result[-1] == inlist[i]
                assert a.get_string() == exlist[i]
                c = []
                result.append(a.string_compare_and_set_value("a", inlist[i]))
                assert result[-1] == exlist[i]
                assert a.get_string() == exlist[i]

                result.append(a.string_compare_and_set_value(inlist[i], exlist[i]))
                assert result[-1] == exlist[i]
                assert a.get_string() == inlist[i]

                value = a.get_string()
                a.reencode('utf-16-le')
                assert a.value == value

                value = a.get_string()
                a.change_mode('m')
                assert value == a.get_string()
                assert a.mode == 'm'
                a.change_mode('s')
                assert value == a.get_string()
                assert a.mode == 's'


            i += 1
    except Exception as e:
        traceback.print_exc()
        print(i)
        raise e


def thread_run(a,c):
    if a.string_compare_and_set_value('cd','ab') == 'ab':
        c.int_add_and_fetch(1)

def test_thread_atomic():
    """
    test single process multiple threads
    :return: None
    """
    a = atomic_string('ab')
    b = atomic_int(0)

    threadlist=[]

    for i in range(10000):
        threadlist.append(Thread(target=thread_run, args=(a,b)))

    for i in range(10000):
        threadlist[i].start()

    for i in range(10000):
        threadlist[i].join()

    assert a.value == 'cd'
    assert b.value == 1


def process_run(a,c):
        from shared_atomic.atomic_string import atomic_string
        from shared_atomic.atomic_int import atomic_int
        def subthread_run(a: atomic_string,c: atomic_int):
            if a.string_compare_and_set_value('cd', 'ab') == 'ab':
                c.int_add_and_fetch(1)

        threadlist = []
        for t in range(5000):
            threadlist.append(Thread(target=subthread_run, args=(a,c)))

        for t in range(5000):
            threadlist[t].start()

        for t in range(5000):
            threadlist[t].join()


def test_process_atomic():
        """
        test multiple processes
        :return: None
        """
        a = atomic_string('ab', mode='m', windows_unix_compatibility=True)
        c = atomic_int(0, mode='m', windows_unix_compatibility=True)
        processlist = []
        for i in range(2):
            processlist.append(Process(target=process_run, args=(a,c)))

        for i in range(2):
            processlist[i].start()

        for i in range(2):
            processlist[i].join()

        assert a.value == 'cd'
        assert c.value == 1

if sys.platform == 'win32':

    from shared_atomic.atomic_object import *

    def process_run_incompatible(a,c):

        def subthread_run(string, reference, integer_reference):
            if string_compare_and_set_value(string, reference, 'cd', 'ab') == 'ab':
                int_add_and_fetch(integer_reference, 1)

        reference_a = get_reference(a)
        reference_c = get_reference(c)
        threadlist = []
        for t in range(5000):
            threadlist.append(Thread(target=subthread_run, args=(a, reference_a.reference, reference_c.reference,)))

        for t in range(5000):
            threadlist[t].start()

        for t in range(5000):
            threadlist[t].join()

        release_reference(reference_a)
        release_reference(reference_c)


    def test_process_atomic_incompatible():
        """
        test multiple processes
        :return: None
        """
        a = atomic_string('ab', mode='m')
        c = atomic_int(0, mode='m')
        processlist = []

        for i in range(10):
            processlist.append(Process(target=process_run_incompatible, args=(a,c,)))

        for i in range(10):
            processlist[i].start()

        for i in range(10):
            processlist[i].join()

        assert a.get_string() == 'cd'
        assert c.get() == 1

