from shared_atomic.atomic_bytearray import atomic_bytearray
from shared_atomic.atomic_shared_memory import atomic_shared_memory
from shared_atomic.atomic_object import *
import sys
from threading import Thread
from multiprocessing import Process
import multiprocessing

if sys.platform != 'win32':
    from shared_atomic.atomic_uint import atomic_uint

if sys.platform != 'win32':
    Process = multiprocessing.get_context("fork").Process


def setup_function():
    """
    pre function for pytest
    :return: None
    """
    # if sys.platform in ('darwin','linux'):
    #     dlltype = ctypes.CDLL
    #     os.chdir('/Users/philren/.local/share/virtualenvs/spark-examples--HrH57AW/lib/python3.6/site-packages')
    #     filename = 'shared_atomic.cpython-36m-darwin.so'
    # elif sys.platform == "win32":
    #     dlltype = ctypes.windll
    # else:
    #     return
def teardown_function():
    pass

def test_init():
    a = atomic_shared_memory(length=1024**3)
    assert a.offset_get(length=1) == b'\0'
    a = atomic_shared_memory(b'1024', mode='m', paddingdirection='l', length=1024**2)
    assert a.offset_get(offset=1024**2-4, length=4) == b'1024'

    if sys.platform in ('darwin', 'linux'):
        a = atomic_shared_memory(b'1024'*1024, mode='m',trimming_direction='l', length=1024*2-1)
        assert a.offset_get(offset=1024*2-5, length=4) == b'1024'
        a = atomic_shared_memory(b'1024'*1024, trimming_direction='r', length=1024*2+1)
        assert a.offset_get(offset=1024*2-7, length=8) == b'02410241'
        assert a.offset_get(offset=1024*2-7, length=2) == b'02'
        assert a.offset_get(offset=1024*2-6, length=1) == b'2'

def test_shared_memory():

    if sys.platform in ('darwin',  'linux'):
        a = atomic_shared_memory(b'1024'*1024, mode='m', trimming_direction='r', length=1024*2+1)
        b = atomic_shared_memory(b'1024'*1024, trimming_direction='r', length=1024*2+1)
        a.offset_store(b, 2, 1, 1)
        r"""a=b'1004'
            b=b'1024'"""
        assert a.offset_get(offset=0, length=4) == b'1004'
        result = a.offset_compare_and_set(b, b'40', 2, 2)
        r"""a=b'1004'
            b=b'1004'"""
        assert result == False
        assert b.offset_get(offset=0, length=4) == b'1004'
        result = a.offset_compare_and_set(b, b'40', 2, 2)
        r"""a=b'1040'
            b=b'1004'"""
        assert result == True
        assert a.offset_get(offset=2, length=2) == b'40'
        r"""a=b'1040'"""
        result = a.offset_add_and_fetch(b'\xFF\xFF', offset=1)
        expected = int.to_bytes(
            (int.from_bytes(b'\xFF\xFF', byteorder='big', signed=False) + \
            int.from_bytes(b'04', byteorder='big', signed=False)) & int('FFFF',base=16),
            length=2, byteorder='big')
        assert result == expected
        assert a.offset_get(offset=1, length=2) == expected

        result = a.offset_sub_and_fetch(b'\xFF\xFF', offset=1)
        expected = b'04'
        assert result == expected
        assert a.offset_get(offset=1, length=2) == expected

        a.offset_memmove(b.offset_get(0,4), offset=0)
        r"""a=b'1004'
            b=b'1004'"""
        assert a.offset_get(offset=0, length=4) == b'1004'
        c = atomic_bytearray(b'1014')
        result = a.offset_compare_with_pointer_and_set(c.reference, b'1024', offset=0)
        assert result == False
        assert c.value == b'1004'
        result = a.offset_compare_with_pointer_and_set(c.reference, b'1024', offset=0)
        assert result == True
        assert a.offset_get(0, 4) == b'1024'
        r"""a=b'1024'
            b=b'1004'"""

        result = a.offset_and_and_fetch(b'2048', offset=4)
        expected = (int.from_bytes(b'1024', byteorder='big') & \
                                                   int.from_bytes(b'2048', byteorder='big', signed=False)).to_bytes(length=4, byteorder='big')
        assert result == expected
        assert a.offset_get(4, 4) == expected

        result = a.offset_or_and_fetch(b'2148', offset=4)
        expected = int.from_bytes(expected, byteorder='big')
        expected |= int.from_bytes(b'2148', byteorder='big')
        expected = expected.to_bytes(length=4, byteorder='big')
        assert result == expected
        assert a.offset_get(4, 4) == expected

        result = a.offset_xor_and_fetch(b'2248', offset=4)
        expected = int.from_bytes(expected, byteorder='big')
        expected ^= int.from_bytes(b'2248', byteorder='big')
        expected = expected.to_bytes(length=4, byteorder='big')
        assert result == expected
        assert a.offset_get(4, 4) == expected

        result = a.offset_nand_and_fetch(b'2348', offset=4)
        expected = int.from_bytes(expected, byteorder='big')
        expected &= int.from_bytes(b'2348', byteorder='big')
        expected = ~expected
        expected = expected.to_bytes(length=4, byteorder='big', signed=True)

        assert result == expected
        assert a.offset_get(4, 4) == expected

        print(a.offset_get(0, 8))
        print(b.offset_get(0, 8))
        a.offset_memmove(b.offset_get(0,4), offset=0)
        r"""a=b'1004'
            b=b'1004'"""
        assert a.offset_get(offset=0, length=8) == b'1004\xff\xfc\xff\xff'
        a.offset_memmove(b'1024', offset=4)

        previous = b'00'
        result = a.offset_fetch_and_add(b'\xFF\xFF', offset=1)
        expected = ((int.from_bytes(b'\xFF\xFF', byteorder='big') + \
                   int.from_bytes(b'00', byteorder='big')) & int('FFFF',base=16)).to_bytes(length=2, byteorder='big')
        assert result == previous
        assert a.offset_get(offset=1, length=2) == expected

        previous = expected
        result = a.offset_fetch_and_sub(b'\xFF\xFF', offset=1)
        expected = int.from_bytes(b'00', byteorder='big').to_bytes(length=2, byteorder='big')
        assert result == previous
        assert a.offset_get(offset=1, length=2) == expected

        previous = b'1024'
        result = a.offset_fetch_and_and(b'2048', offset=4)
        expected = int.from_bytes(previous, byteorder='big')
        expected &= int.from_bytes(b'2048', byteorder='big')
        expected = expected.to_bytes(length=4, byteorder='big')
        assert result == previous
        assert a.offset_get(4, 4) == expected

        previous = expected
        result = a.offset_fetch_and_or(b'2148', offset=4)
        expected = int.from_bytes(expected, byteorder='big')
        expected |= int.from_bytes(b'2148', byteorder='big')
        expected = expected.to_bytes(length=4, byteorder='big')
        assert result == previous
        assert a.offset_get(4, 4) == expected

        previous = expected
        result = a.offset_fetch_and_xor(b'2248', offset=4)
        expected = int.from_bytes(expected, byteorder='big')
        expected ^= int.from_bytes(b'2248', byteorder='big')
        expected = expected.to_bytes(length=4, byteorder='big')
        assert result == previous
        assert a.offset_get(4, 4) == expected

        previous = expected
        result = a.offset_fetch_and_nand(b'2348', offset=4)
        expected = int.from_bytes(expected, byteorder='big')
        expected &= int.from_bytes(b'2348', byteorder='big')
        expected = ~expected
        expected = expected.to_bytes(length=4, byteorder='big', signed=True)

        assert result == previous
        assert a.offset_get(4, 4) == expected

        c = atomic_uint(2**64-1)
        a.offset_store_from_pointer(c.reference, 0)

        assert a.offset_get(0, 8) == int.to_bytes(2**64-1, 8, 'big')

        previous = a.offset_get(0,8)
        a.change_mode('s')
        assert a.mode == 's'
        assert previous == a.offset_get(0,8)
        a.change_mode('m')
        assert a.mode == 'm'
        assert previous == a.offset_get(0,8)


        a = atomic_shared_memory(b'1024'*1024, mode='m', trimming_direction='r', length=1024*2+1)
        result = a.offset_compare_and_set_value(b'04', b'04', 2)
        assert result == b'24'
        assert a.offset_get(0, 4) == b'1024'
        result = a.offset_compare_and_set_value(b'04', b'24', 2)
        assert result == b'24'
        assert a.offset_get(0, 4) == b'1004'

    else:
        a = atomic_shared_memory(b'1024' * 1024, mode='m', trimming_direction='r', length=1024 * 2 + 1)
        b = atomic_shared_memory(b'1024' * 1024, trimming_direction='l', length=1024 * 2 + 1)

        r"""a=b'102410241'
            b=b'410241024'"""

        assert a.offset_get(offset=0, length=4) == b'1024'
        a.offset_get_and_set(b'1004',4)
        r"""a=b'100410241'
            b=b'410241024'"""
        assert a.offset_get(offset=4, length=4) == b'1004'
        a.offset_memmove(b.offset_get(1,1),1)
        assert a.offset_get(offset=0, length=4) == b'1124'
        r"""a=b'110410241'
            b=b'410241024'"""
        a = atomic_shared_memory(b'1024' * 1024, mode='m', trimming_direction='r', length=1024 * 2 + 1)
        result = a.offset_add_and_fetch(b'\xFF\xFF\xFF\xFF', offset=4)
        expected = int.to_bytes(
            (int.from_bytes(b'\xFF\xFF\xFF\xFF', byteorder='big', signed=True) + \
             int.from_bytes(b'1024', byteorder='big', signed=True)) & int('FFFFFFFF', base=16),
            length=4, byteorder='big')
        assert result == expected
        assert a.offset_get(offset=4, length=4) == expected

        result = a.offset_sub_and_fetch(b'\xFF\xFF\xFF\xFF', offset=4)
        expected = b'1024'
        assert result == expected
        assert a.offset_get(offset=4, length=4) == expected
        r"""a=b'102410241'
            b=b'410241024'"""
        result = a.offset_compare_and_set_value(b"ab", b"11", 4)
        assert a.offset_get(offset=4, length=4) == b'1024'
        assert result == b'10'
        result = a.offset_compare_and_set_value(b"ab", b"10", 4)
        assert a.offset_get(offset=4, length=4) == b'ab24'
        assert result == b'10'
        r"""a=b'11ab10241'
            b=b'410241024'"""
        a = atomic_shared_memory(b'11ab10241', mode='m')

        previous = b'10'
        result = a.offset_fetch_and_add(b'\xFF\xFF', offset=4)
        expected = ((int.from_bytes(b'\xFF\xFF', byteorder='big') + \
                     int.from_bytes(b'10', byteorder='big')) & int('FFFF', base=16)).to_bytes(length=2, byteorder='big')
        assert result == previous
        assert a.offset_get(offset=4, length=2) == expected

        previous = expected
        result = a.offset_fetch_and_sub(b'\xFF\xFF', offset=4)
        expected = int.from_bytes(b'10', byteorder='big').to_bytes(length=2, byteorder='big')
        assert result == previous
        assert a.offset_get(offset=4, length=2) == expected

        previous = b'1024'
        result = a.offset_fetch_and_and(b'2048', offset=4)
        expected = int.from_bytes(previous, byteorder='big')
        expected &= int.from_bytes(b'2048', byteorder='big')
        expected = expected.to_bytes(length=4, byteorder='big')
        assert result == previous
        assert a.offset_get(4, 4) == expected

        previous = expected
        result = a.offset_fetch_and_or(b'2148', offset=4)
        expected = int.from_bytes(expected, byteorder='big')
        expected |= int.from_bytes(b'2148', byteorder='big')
        expected = expected.to_bytes(length=4, byteorder='big')
        assert result == previous
        assert a.offset_get(4, 4) == expected

        previous = expected
        result = a.offset_fetch_and_xor(b'2248', offset=4)
        expected = int.from_bytes(expected, byteorder='big')
        expected ^= int.from_bytes(b'2248', byteorder='big')
        expected = expected.to_bytes(length=4, byteorder='big')
        assert result == previous
        assert a.offset_get(4, 4) == expected

        previous = a.offset_get(0, 8)
        a.change_mode('s')
        assert a.mode == 's'
        assert previous == a.offset_get(0, 8)
        a.change_mode('m')
        assert a.mode == 'm'
        assert previous == a.offset_get(0, 8)

        a = atomic_shared_memory(b'1024' * 1024)
        assert a.offset_bittest_and_reset(35) == True
        assert '0' * (64 - len(bin(int.from_bytes(a.offset_get(0, 8), byteorder='big'))[2:])) + bin(
            int.from_bytes(a.offset_get(0, 8), byteorder='big'))[2:] \
               == '0011000100110000001100100011010000100001001100000011001000110100'
        assert a.offset_bittest_and_set(35) == False
        assert '0'*(64-len(bin(int.from_bytes(a.offset_get(0,8), byteorder='big'))[2:]))+bin(int.from_bytes(a.offset_get(0,8), byteorder='big'))[2:] \
         == '0011000100110000001100100011010000110001001100000011001000110100'
    

def thread_run(a,):
    if a.offset_compare_and_set_value(b'bc', b'cd', 2) == b'bc':
        assert a.offset_get(0, 4) == b'abbc'

def test_thread_atomic():
    """
    test single process multiple threads
    :return: None
    """
    a = atomic_shared_memory(b'abcd', mode='s', length=1024 ** 3)
    threadlist = []
    for i in range(10000):
        threadlist.append(Thread(target=thread_run, args=(a,)))
    for i in range(10000):
        threadlist[i].start()
    for i in range(10000):
        threadlist[i].join()
    assert a.offset_get(0,2) == b'ab'


def process_run_compatibility(a: atomic_shared_memory):
    from shared_atomic.atomic_shared_memory import atomic_shared_memory
    def subthread_run(a: atomic_shared_memory):
        if a.offset_compare_and_set_value(b'bc',b'cd',2) == b'cd':
            assert a.offset_get(0,4) == b'abbc'

    threadlist = []
    for t in range(5000):
        threadlist.append(Thread(target=subthread_run, args=(a,)))

    for t in range(5000):
        threadlist[t].start()

    for t in range(5000):
        threadlist[t].join()


def test_process_atomic_compatibility():
    """
    test multiple processes
    :return: None
    """
    a = atomic_shared_memory(b'abcd', mode='m', length=1024 ** 3, windows_unix_compatibility=True)

    processlist = []
    for i in range(2):
        processlist.append(Process(target=process_run_compatibility, args=(a,)))

    for i in range(2):
        processlist[i].start()

    for i in range(2):
        processlist[i].join()

    assert a.offset_get(0, 4) == b'abbc'

if sys.platform == 'win32':

    def process_run(a):
        def subthread_run(a, reference):
            if shared_memory_offset_compare_and_set_value(a, reference, b'bc', b'cd', 2) == b'cd':
                assert shared_memory_offset_get(a, reference, 0, 4) == b'abbc'

        reference = get_reference(a)

        threadlist = []
        for t in range(50000):
            threadlist.append(Thread(target=subthread_run, args=(a, reference.reference,)))
        for t in range(50000):
            threadlist[t].start()
        for t in range(50000):
            threadlist[t].join()

        release_reference(reference)


    def test_process_atomic_incompatible():
        """
        test multiple processes
        :return: None
        """
        a = atomic_shared_memory(b'abcd', mode='m', length=1024 ** 3, windows_unix_compatibility=False)
        processlist = []

        for i in range(10):
            processlist.append(Process(target=process_run, args=(a,)))

        for i in range(10):
            processlist[i].start()

        for i in range(10):
            processlist[i].join()

        assert a.offset_get(0, 4) == b'abbc'